from .backend import Backend
from qutrunk.sim.local.service import BackendLocalService
from qutrunk.tools.read_qubox import get_qulocalbox_setting


class BackendLocal(Backend):
    """Local simulator simulates a quantum compute using C++-based kernels."""
    def __init__(self):
        super().__init__()
        self.circuit = None
        self._local_server = BackendLocalService()
        box_config = get_qulocalbox_setting()
        self._show_quantum_gate = box_config.get('show_quantum_gate')

    def send_circuit(self, circuit, final=False):
        start = circuit.cmd_cursor
        stop = len(circuit.cmds)

        if start == 0:
            res, elapsed = self._local_server.init(
                len(circuit.qreg), 
                circuit.initstate, 
                circuit.initvalue,
                self._show_quantum_gate)
            if self.circuit.counter:
                self.circuit.counter.acc_run_time(elapsed)

        res, elapsed = self._local_server.send_circuit(circuit, final)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)

        circuit.forward(stop - start)

        return res

    def run(self, shots = 1):
        res, elapsed = self._local_server.run(shots)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
            self.circuit.counter.finish() 
        return res

    def get_prob_amp(self, qubit):
        """
        Get the probability of a state-vector at an index in the full state vector
        """
        res, elapsed = self._local_server.get_prob_amp(qubit)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def get_prob_outcome(self, qubit, outcome):
        """
        Gives the probability of a specified qubit being measured in the given outcome (0 or 1)
        """
        res, elapsed = self._local_server.get_prob_outcome(qubit, outcome)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def get_prob_all_outcome(self, qubits):
        """
        Populates outcomeProbs with the probabilities of every outcome of the sub-register contained in qubits
        """
        res, elapsed = self._local_server.get_prob_all_outcome(qubits)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def get_all_state(self):
        """
        Get the current state vector of probability amplitudes for a set of qubits
        """
        res, elapsed = self._local_server.get_all_state()
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def apply_QFT(self, qubits):
        res, elapsed = self._local_server.apply_QFT(qubits)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def apply_Full_QFT(self):
        res, elapsed = self._local_server.apply_Full_QFT()
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res
