from qutrunk.sim.qusim.rpcclient import QuSimApiServer
from qutrunk.sim.qusim.qusim import ttypes as qusim
from qutrunk.tools.read_qubox import get_qubox_setting
from .backend import Backend


class BackendQuSim(Backend):
    """
    quroot:
    The qudoor quantum simulator,
    use classical architecture to simulate the quantum computing environment,
    used for running quantum circuits

    BackendQuRoot:
    Establish network communication with QuRoot and transfer the quantum circuit to QuRoot for operation

    Backend:All simulators are derived from this class.
    """
    def __init__(self):
        super().__init__()
        self.circuit = None
        box_config = get_qubox_setting()
        self._api_server = QuSimApiServer(ip=box_config.get('ip'), port=box_config.get('port'))

    def get_prob_amp(self, index):
        """
        Get the probability of a state-vector at an index in the full state vector
        """
        res, elapsed = self._api_server.get_prob_amp(index)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def get_prob_all_outcome(self, qubits):
        """
        Populates outcomeProbs with the probabilities of every outcome of the sub-register contained in qubits
        """
        res, elapsed = self._api_server.get_prob_all_outcome(qubits)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def get_prob_outcome(self, qubit, outcome):
        """
        Gives the probability of a specified qubit being measured in the given outcome (0 or 1)
        """
        res, elapsed = self._api_server.get_prob_outcome(qubit, outcome)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def get_all_state(self):
        """
        Get the current state vector of probability amplitudes for a set of qubits
        """
        res, elapsed = self._api_server.get_all_state()
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def send_circuit(self, circuit, final=False):
        """
        send the quantum circuit to qubox through rpc
        """
        cmds = []
        start = circuit.cmd_cursor
        stop = len(circuit.cmds)
        for idx in range(start, stop):
            cmd = circuit.cmds[idx]
            c = qusim.Cmd(str(cmd.gate), cmd.targets, cmd.controls, cmd.rotation, cmd.qasm)
            cmds.append(c)

        circuit.forward(stop - start)

        # 服务端初始化
        if start == 0:
            res, elapsed = self._api_server.init(
                circuit.qubits_len, 
                circuit.initstate, 
                circuit.initvalue,
                circuit.get_density())
            if self.circuit.counter:
                self.circuit.counter.acc_run_time(elapsed)

        if len(cmds) == 0 and (not final):
            return

        # 发送至服务        
        res, elapsed = self._api_server.send_circuit(qusim.Circuit(cmds), final)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)

    def run(self, shots = 1):
        res, elapsed = self._api_server.run(shots)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
            self.circuit.counter.finish() 

        """
        1 必须释放连接，不然其它连接无法连上服务端
        2 不能放在__del__中，因为对象释放不代表析构函数会及时调用
        """
        self._api_server.close()

        return res

    def apply_QFT(self, qubits):
        self._api_server.apply_QFT(qubits)

    def apply_Full_QFT(self):
        self._api_server.apply_Full_QFT()
