from enum import Enum
import json

from qutrunk.core.qureg import Qureg
from qutrunk.core.qubit import Qubit
from qutrunk.core.backends import BackendLocal
from qutrunk.core.backends import BackendQuSim
from qutrunk.core.backends import Backend


class InitState(Enum):
    """
    Init state for quantum circuit(represent by qureg)

    Attributes:
        Zero: Initialise qureg into the zero state
        Plus: Initialise qureg into the plus state.
        Classical: Initialise qureg into the classical state with index stateInd
    """
    Zero = 1
    Plus = 2
    Classical = 3


class QCircuit:
    """
    Quantum circuit

    Attributes:
        backend: Used to run quantum circuits
        density: Creates a density matrix Qureg object representing a set of qubits which can enter noisy and mixed states
    """
    def __init__(self, backend=None, density=False):
        self.qreg = Qureg()
        self.creg = []
        self.cmds = []
        self.init_state = None
        self.init_value = None
        # 是否启用噪声模型
        self.density = density
        self.cmd_cursor = 0
        self.counter = None

        # 默认使用local作为后端
        if backend is None:
            self.backend = BackendLocal()
        else:
            if not isinstance(backend, Backend):
                raise TypeError("You supplied a backend which is not supported.\n")
            self.backend = backend

        self.backend.circuit = self
        self.outcome = None

    def __iter__(self):
        """
        Used to iterate commands in quantum circuits
        """
        return QCircuitIter(self.cmds)

    def allocate(self, qubits, init_state=InitState.Zero, init_value=0):
        """
        Allocate qubit in quantum circuit
        Args:
            qubits: number of qubit allocated in circuit
            initstate: Zero, Plus or Classical state
        """
        self.init_state = init_state
        self.init_value = init_value
        for id in range(qubits):
            self.qreg.append(Qubit(self, id))
            # 默认为：-1，这样可以区分量子比特是否被测量过
            self.creg.append(-1)
        return self.qreg

    def get_init_qasm(self):
        """Convert the initialization state of the circuit into qasm instruction"""
        qasm = ""
        if self.init_state == InitState.Plus:
            qasm += "reset q;\n"
            qasm += "h q;\n"
        elif self.init_state == InitState.Classical:
            binstr = "{0:b}".format(self.init_value)
            binarr = [int(b) for b in binstr]
            binarr = binarr[::-1]
            qasm += "reset q;\n"
            for i, val in enumerate(binarr):
                if val == 1:
                    qasm += "x q[" + str(i) + "];\n"
        return qasm

    def get_init_quqasm(self):
        pass

    def set_cmds(self, cmds):
        self.cmds = cmds

    def append_cmd(self, cmd):
        """
        Append command to circuit when apply a quantum gate
        Args:
            cmd: the command to append
        """
        self.cmds.append(cmd)

    @property
    def initstate(self):
        """
        Get circuit init state
        """
        return self.init_state.value

    @property
    def initvalue(self):
        """
        Get circuit init value
        """
        return self.init_value 
    
    def get_density(self):
        """
        Get circuit density
        """
        return self.density

    def forward(self, num):
        """
        Update the cmd_cursor when a bunch of quantum operations have been run
        """
        self.cmd_cursor += num
    
    @property
    def qubits_len(self):
        """
        Get the number of qubits
        """
        return len(self.qreg)

    @property
    def gates_len(self):
        """
        Get the number of gates
        """
        return len(self.cmds) 

    def set_measure(self, qubit, value):
        """Store the measure result for target qubit"""
        if qubit >= len(self.qreg) or qubit < 0:
            raise Exception('量子比特下标越界') 
        self.creg[qubit] = value

    # TODO: 如果用户没有主动调run结束程序，那么资源的释放应该放到析构函数里面
    def run(self, shots = 1):
        self.backend.send_circuit(self, True)
        result = self.backend.run(shots)
        if result and result.measureSet:
            for m in result.measureSet:
                self.set_measure(m.id, m.value)

        res = Result(self.qubits_len, result)

        return res

    def apply_QFT(self, qubits):
        """
        Applies the quantum Fourier transform (QFT) to a specific subset of qubits of the register qureg

        Args:
            qubits: a list of the qubits to operate the QFT upon
        """
        self.backend.send_circuit(self)
        self.backend.apply_QFT(qubits)

    def apply_Full_QFT(self):
        """
        Applies the quantum Fourier transform (QFT) to the entirety of qureg
        """
        self.backend.send_circuit(self)
        self.backend.apply_Full_QFT()

    def get_prob_amp(self, index):
        """
        Get the probability of a state-vector at an index in the full state vector.

        Args:
            index: index in state vector of probability amplitudes
        """
        self.backend.send_circuit(self)
        prob = self.backend.get_prob_amp(index)
        return prob

    def get_prob_outcome(self, qubit, outcome):
        """
        Gives the probability of a specified qubit being measured in the given outcome (0 or 1).

        Args:
            qubit: qubit to study
            outcome: for which to find the probability of the qubit being measured in
        """
        self.backend.send_circuit(self)
        prob = self.backend.get_prob_outcome(qubit, outcome)
        return prob 

    def get_prob_all_outcome(self, qubits):
        """
        Populates outcomeProbs with the probabilities of every outcome of the sub-register contained in qubits

        Args:
            qubits: a list of qubits to study
        """
        self.backend.send_circuit(self)
        probs = self.backend.get_prob_all_outcome(qubits)
        return probs

    def get_all_state(self):
        """
        Get the current state vector of probability amplitudes for a set of qubits
        """
        self.backend.send_circuit(self)
        state = self.backend.get_all_state()
        return state

class QCircuitIter:
    """The iterator for circuit"""
    def __init__(self, cmds):
        self.idx = 0
        self.cmds = cmds

    def __iter__(self):
        return self

    def __next__(self):
        if self.idx < len(self.cmds):
            idx = self.idx
            self.idx += 1
            return self.cmds[idx]
        else:
            raise StopIteration


class Result:
    """save the result of quantum circuit running"""
    def __init__(self, num_qubits, res):
        # 每个量子比特的测量结果
        self.measure_result = [-1] * num_qubits
        if res and res.measureSet:
            for m in res.measureSet:
                self.set_measure(m.id, m.value)
        # 统计所有量子比特组成的比特位字符串出现的次数
        self.outcome = res.outcomeSet

    def set_measure(self, qubit, value):
        if qubit >= len(self.measure_result) or qubit < 0:
            raise Exception("量子比特下标越界了")
        self.measure_result[qubit] = value

    def get_measure(self):
        """Get the measure result"""
        return self.measure_result

    def get_outcome(self):
        """Get the measure result in binary format"""
        out = self.measure_result[::-1]
        bitstr = "0b"
        for o in out:
            bitstr += str(o)
        return bitstr

    def get_counts(self):
        """Get the number of times the measurement results appear"""
        if self.outcome is None:
            return
            
        res = []
        for out in self.outcome:
            temp = {out.bitstr: out.count}
            res.append(temp)
        return json.dumps(res)
