import numpy as np
import cmath
import math


from qutrunk.core.command import Command
from qutrunk.core.qubit import Qubit


class BasicGate:
    def __init__(self):
        pass

    def __str__(self):
        '''The string description of a quantum logic gate'''
        return ""

    def __or__(self, qubit):
        '''Overwrite | operator to achieve quantum logic gate operation'''
        pass

    def qasm_str(self, qubit):
        '''Qasm string used to describe quantum logic gates'''
        pass


class U3(BasicGate):
    '''
    U3 gate
    '''
    def __init__(self, alpha, beta, theta):
        '''
        Args:
            alpha: 
            beta:
            theta:
        '''
        super().__init__()
        self.alpha = alpha
        self.beta = beta
        self.theta = theta

    def __str__(self):
        return "U3"

    def quqasm_str(self, qubit):
        return f"{self.__str__()}({str(self.alpha)}, {str(self.beta)}, { str(self.theta)}) | q[" + str(qubit.id) + "]"

    def qasm_str(self, qubit):
        return "u3(" + str(self.alpha) + "," + str(self.beta) + "," + str(self.theta) + ") q[" + str(qubit.id) + "]"

    def __or__(self, qubit):
        '''
        Args:
            qubit: the quantum bit to aplly U3 gate
        Example: U3(pi/2,pi/2,pi/2) | qr[0]
        ''' 
        targets = [qubit.id]
        cmd = Command(self, targets, rotation=[self.alpha, self.beta, self.theta], qasm=self.qasm_str(qubit), quqasm=self.quqasm_str(qubit))
        qubit.circuit.append_cmd(cmd)  

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        theta = self.alpha
        phi = self.beta
        lam = self.theta
        cos = np.cos(theta / 2)
        sin = np.sin(theta / 2)
        return np.array(
            [
                [cos, -np.exp(1j * lam) * sin],
                [np.exp(1j * phi) * sin, np.exp(1j * (phi + lam)) * cos],
            ]
        )

class U(U3):
    pass
    
class U2(BasicGate):
    '''
    U2 gate
    '''
    def __init__(self, alpha, beta):
        '''
        Args:
            alpha: 
            beta:
        '''
        super().__init__()
        self.alpha = alpha
        self.beta = beta

    def __str__(self):
        return "U2"

    def quqasm_str(self, qubit):
        return f"{self.__str__()}({str(self.alpha)}, {str(self.beta)}) | q[" + str(qubit.id) + "]"

    def qasm_str(self, qubit):
        return "u2(" + str(self.alpha) + "," + str(self.beta) + ") q[" + str(qubit.id) + "]"

    def __or__(self, qubit):
        '''
        Args:
            qubit: the quantum bit to aplly U2 gate
        Example: U2(pi/2,pi/2) | qr[0]
        ''' 
        targets = [qubit.id]
        cmd = Command(self, targets, rotation=[self.alpha, self.beta], qasm=self.qasm_str(qubit), quqasm=self.quqasm_str(qubit))
        qubit.circuit.append_cmd(cmd)  

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        isqrt2 = 1 / np.sqrt(2)
        phi = self.alpha
        lam = self.beta
        return np.matrix(
            [
                [isqrt2, -np.exp(1j * lam) * isqrt2],
                [np.exp(1j * phi) * isqrt2, np.exp(1j * (phi + lam)) * isqrt2],
            ]
        )

class U1(BasicGate):
    '''
    U1 gate
    '''
    def __init__(self, alpha):
        '''
        Args:
            alpha: 
        '''
        super().__init__()
        self.alpha = alpha

    def __str__(self):
        return "U1"

    def quqasm_str(self, qubit):
        return f"{self.__str__()}({str(self.alpha)}) | q[" + str(qubit.id) + "]"

    def qasm_str(self, qubit):
        return "u1(" + str(self.alpha) + ") q[" + str(qubit.id) + "]"

    def __or__(self, qubit):
        '''
        Args:
            qubit: the quantum bit to aplly U1 gate
        Example: U1(pi/2) | qr[0]
        ''' 
        targets = [qubit.id]
        cmd = Command(self, targets, rotation=[self.alpha], qasm=self.qasm_str(qubit), quqasm=self.quqasm_str(qubit))
        qubit.circuit.append_cmd(cmd)  

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        lam = float(self.alpha)
        return np.matrix([[1, 0], [0, np.exp(1j * lam)]])

class XGate(BasicGate):
    '''Apply the single-qubit Pauli-X (also known as the X, sigma-X, NOT or bit-flip) gate'''
    def __init__(self):
        super().__init__()

    def __str__(self):
        return "X"

    def quqasm_str(self, qubit):
        return f"{self.__str__()} | q[" + str(qubit.id) + "]"

    def qasm_str(self, qubit):
        return "x q[" + str(qubit.id) + "]"

    def __or__(self, qubit):
        '''
        Args:
            qubit: the quantum bit to aplly X gate
        Example: X | qr[0]
        ''' 
        targets = [qubit.id]
        cmd = Command(self, targets, qasm=self.qasm_str(qubit), quqasm=self.quqasm_str(qubit))
        qubit.circuit.append_cmd(cmd) 

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix([[0, 1], [1, 0]])

X = NOT = XGate()

class YGate(BasicGate):
    '''Apply the single-qubit Pauli-Y (also known as the Y or sigma-Y) gate'''
    def __init__(self):
        super().__init__()

    def __str__(self):
        return "Y"

    def quqasm_str(self, qubit):
        return f"{self.__str__()} | q[" + str(qubit.id) + "]"

    def qasm_str(self, qubit):
        return "y q[" + str(qubit.id) + "]"

    def __or__(self, qubit):
        '''
        Args:
            qubit: the quantum bit to aplly Y gate
        Example: Y | qr[0]
        ''' 
        targets = [qubit.id]
        cmd = Command(self, targets, qasm=self.qasm_str(qubit), quqasm=self.quqasm_str(qubit))
        qubit.circuit.append_cmd(cmd) 

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix([[0, -1j], [1j, 0]])

Y = YGate()


class ZGate(BasicGate):
    '''Apply the single-qubit Pauli-Z (also known as the Z, sigma-Z or phase-flip) gate'''
    def __init__(self):
        super().__init__()

    def __str__(self):
        return "Z"

    def quqasm_str(self, qubit):
        return f"{self.__str__()} | q[" + str(qubit.id) + "]"

    def qasm_str(self, qubit):
        return "z q[" + str(qubit.id) + "]"

    def __or__(self, qubit):
        '''
        Args:
            qubit: the quantum bit to aplly Z gate
        Example: Z | qr[0]
        ''' 
        targets = [qubit.id]
        cmd = Command(self, targets, qasm=self.qasm_str(qubit), quqasm=self.quqasm_str(qubit))
        qubit.circuit.append_cmd(cmd) 

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix([[1, 0], [0, -1]])


Z = ZGate()


class HGate(BasicGate):
    '''
    Apply the single-qubit Hadamard gate
    '''
    def __init__(self):
        super().__init__()

    def __str__(self):
        return "H"

    def quqasm_str(self, qubit):
        return f"{self.__str__()} | q[" + str(qubit.id) + "]"

    def qasm_str(self, qubit):
        return "h q[" + str(qubit.id) + "]"

    def __or__(self, qubit):
        '''
        Overwrite | operator to achieve quantum logic gate operation
        Args:
            qubit: quantum bit applied to hadamard gate
        Example: H | qr[0]
        '''
        targets = [qubit.id]
        # TODO:传参？
        cmd = Command(self, targets, qasm=self.qasm_str(qubit), quqasm=self.quqasm_str(qubit))
        qubit.circuit.append_cmd(cmd)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return 1.0 / cmath.sqrt(2.0) * np.matrix([[1, 1], [1, -1]])


H = HGate()


class MeasureGate(BasicGate):
    '''Measures a single qubit, collapsing it randomly to 0 or 1.'''
    def __init__(self):
        super().__init__()

    def __str__(self):
        return "Measure"

    def quqasm_str(self, qubit):
        return f"{self.__str__()} | q[" + str(qubit.id) + "]"

    def qasm_str(self, qubit):
        return "measure q[" + str(qubit.id) + "] -> c[" + str(qubit.id) + "]" 

    def __or__(self, qubit):
        '''
        Args:
            qubit: the quantum bit to be measured
        Example: Measure | qr[0]
        '''
        targets = [qubit.id]
        cmd = Command(self, targets, qasm=self.qasm_str(qubit), quqasm=self.quqasm_str(qubit))
        qubit.circuit.append_cmd(cmd)


Measure = MeasureGate()


class BasicRotateGate(BasicGate):
    '''
    Base class for rotation gate
    '''
    def __init__(self):
        super().__init__()


class BasciPhaseGate(BasicGate):
    '''
    Base class for phase gate
    '''
    def __init__(self):
        super().__init__()


class P(BasciPhaseGate):
    '''
    Phase gate (global phase)
    Args:
        alpha: the phase to apply
    '''
    def __init__(self, alpha):
        super().__init__()
        self.rotation = alpha

    def __str__(self):
        return "P"

    def quqasm_str(self, qubit):
        return f"{self.__str__()}(" + str(self.rotation) + ") | q[" + str(qubit.id) + "]"

    def qasm_str(self, qubit):
        return "p (" + str(self.rotation) + ") q[" + str(qubit.id) + "]"

    def __or__(self, qubit):
        '''
        Args:
            qubit: the quantum bit to aplly phase gate
        Example: P(alpha) | qr[0]
        '''
        targets = [qubit.id]
        cmd = Command(self, targets, rotation=[self.rotation], qasm=self.qasm_str(qubit), quqasm=self.quqasm_str(qubit))
        qubit.circuit.append_cmd(cmd) 

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix([[cmath.exp(1j * self.rotation), 0], [0, cmath.exp(1j * self.rotation)]])


class R(BasicRotateGate):
    '''
    Shift the phase between |0> and |1> of a single qubit by a given angle
    Args:
        alpha: the angle to rotate
    '''
    def __init__(self, alpha):
        super().__init__()
        self.rotation = alpha

    def __str__(self):
        return "R"

    def quqasm_str(self, qubit):
        return f"{self.__str__()}(" + str(self.rotation) + ") | q[" + str(qubit.id) + "]"

    def qasm_str(self, qubit):
        return "r (" + str(self.rotation) + ") q[" + str(qubit.id) + "]"

    def __or__(self, qubit):
        '''
        Args:
            qubit: the quantum bit to aplly R gate
        Example: R(alpha) | qr[0]
        ''' 
        targets = [qubit.id]
        cmd = Command(self, targets, rotation=[self.rotation], qasm=self.qasm_str(qubit), quqasm=self.quqasm_str(qubit))
        qubit.circuit.append_cmd(cmd)  

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix([[1, 0], [0, cmath.exp(1j * self.rotation)]])


class Rx(BasicRotateGate):
    '''
    Rotate a single qubit by a given angle around the X-axis of the Bloch-sphere
    Args:
        alpha: the angle to rotate
    '''
    def __init__(self, alpha):
        super().__init__()
        self.rotation = alpha

    def __str__(self):
        return "Rx"

    def quqasm_str(self, qubit):
        return f"{self.__str__()}(" + str(self.rotation) + ") | q[" + str(qubit.id) + "]"

    def qasm_str(self, qubit):
        return "rx (" + str(self.rotation) + ") q[" + str(qubit.id) + "]"

    def __or__(self, qubit):
        '''
        Args:
            qubit: the quantum bit to aplly Rx gate
        Example: Rx(alpha) | qr[0]
        ''' 
        targets = [qubit.id]
        cmd = Command(self, targets, rotation=[self.rotation], qasm=self.qasm_str(qubit), quqasm=self.quqasm_str(qubit))
        qubit.circuit.append_cmd(cmd) 

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix(
            [
                [math.cos(0.5 * self.rotation), -1j * math.sin(0.5 * self.rotation)],
                [-1j * math.sin(0.5 * self.rotation), math.cos(0.5 * self.rotation)],
            ]
        )


class Ry(BasicRotateGate):
    '''
    Rotate a single qubit by a given angle around the Y-axis of the Bloch-sphere
    Args:
        alpha: the angle to rotate
    '''
    def __init__(self, alpha):
        super().__init__()
        self.rotation = alpha

    def __str__(self):
        return "Ry"

    def quqasm_str(self, qubit):
        return f"{self.__str__()}(" + str(self.rotation) + ") | q[" + str(qubit.id) + "]"

    def qasm_str(self, qubit):
        return "ry (" + str(self.rotation) + ") q[" + str(qubit.id) + "]"

    def __or__(self, qubit):
        '''
        Args:
            qubit: the quantum bit to aplly Ry gate
        Example: Ry(alpha) | qr[0]
        ''' 
        targets = [qubit.id]
        cmd = Command(self, targets, rotation=[self.rotation], qasm=self.qasm_str(qubit), quqasm=self.quqasm_str(qubit))
        qubit.circuit.append_cmd(cmd) 

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix(
            [
                [math.cos(0.5 * self.rotation), -math.sin(0.5 * self.rotation)],
                [math.sin(0.5 * self.rotation), math.cos(0.5 * self.rotation)],
            ]
        )


class Rz(BasicRotateGate):
    '''
    Rotate a single qubit by a given angle around the Z-axis of the Bloch-sphere (also known as a phase shift gate).
    Args:
        alpha: the angle to rotate
    '''
    def __init__(self, alpha):
        super().__init__()
        self.rotation = alpha

    def __str__(self):
        return "Rz"

    def quqasm_str(self, qubit):
        return f"{self.__str__()}(" + str(self.rotation) + ") | q[" + str(qubit.id) + "]"

    def qasm_str(self, qubit):
        return "rz (" + str(self.rotation) + ") q[" + str(qubit.id) + "]"

    def __or__(self, qubit):
        '''
        Args:
            qubit: the quantum bit to aplly Rz gate
        Example: Rz(alpha) | qr[0]
        ''' 
        targets = [qubit.id]
        cmd = Command(self, targets, rotation=[self.rotation], qasm=self.qasm_str(qubit), quqasm=self.quqasm_str(qubit))
        qubit.circuit.append_cmd(cmd) 

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix(
            [
                [cmath.exp(-0.5 * 1j * self.rotation), 0],
                [0, cmath.exp(0.5 * 1j * self.rotation)],
            ]
        )


class Rxx(BasicRotateGate):
    '''
    RotationXX gate class
    Args:
        alpha: the angle to rotate
    '''
    def __init__(self, alpha):
        super().__init__()
        self.rotation = alpha

    def __str__(self):
        return "Rxx"

    def quqasm_str(self, qubits):
        return f"{self.__str__()}(" + str(self.rotation) + ") | (q[" + str(qubits[0].id) + "], q[" + str(qubits[1].id) + "])"

    def qasm_str(self, qubits):
        return "rxx (" + str(self.rotation) + ") q[" + str(qubits[0].id) + "],q[" + str(qubits[1].id) + "]"

    def __or__(self, qubits):
        '''
        Args:
            qubits: the quantum bits to aplly Rxx gate
        Example: Rxx(alpha) | (qr[0], qr[1])
        ''' 
        if len(qubits) != 2:
            raise AttributeError()
        targets = [q.id for q in qubits]
        cmd = Command(self, targets, rotation=[self.rotation], qasm=self.qasm_str(qubits), quqasm=self.quqasm_str(qubits))
        qubits[0].circuit.append_cmd(cmd)  

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix(
            [
                [cmath.cos(0.5 * self.rotation), 0, 0, -1j * cmath.sin(0.5 * self.rotation)],
                [0, cmath.cos(0.5 * self.rotation), -1j * cmath.sin(0.5 * self.rotation), 0],
                [0, -1j * cmath.sin(0.5 * self.rotation), cmath.cos(0.5 * self.rotation), 0],
                [-1j * cmath.sin(0.5 * self.rotation), 0, 0, cmath.cos(0.5 * self.rotation)],
            ]
        )


class Ryy(BasicRotateGate):
    '''
    RotationYY gate class
    Args:
        alpha: the angle to rotate
    '''
    def __init__(self, alpha):
        super().__init__()
        self.rotation = alpha

    def __str__(self):
        return "Ryy"

    def quqasm_str(self, qubits):
        return f"{self.__str__()}(" + str(self.rotation) + ") | (q[" + str(qubits[0].id) + "], q[" + str(qubits[1].id) + "])"

    def qasm_str(self, qubits):
        return "ryy (" + str(self.rotation) + ") q[" + str(qubits[0].id) + "],q[" + str(qubits[1].id) + "]"

    def __or__(self, qubits):
        '''
        Args:
            qubits: the quantum bits to aplly Ryy gate
        Example: Ryy(alpha) | (qr[0], qr[1])
        ''' 
        if len(qubits) != 2:
            raise AttributeError()
        targets = [q.id for q in qubits]
        cmd = Command(self, targets, rotation=[self.rotation], qasm=self.qasm_str(qubits), quqasm=self.quqasm_str(qubits))
        qubits[0].circuit.append_cmd(cmd)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix(
            [
                [cmath.cos(0.5 * self.rotation), 0, 0, 1j * cmath.sin(0.5 * self.rotation)],
                [0, cmath.cos(0.5 * self.rotation), -1j * cmath.sin(0.5 * self.rotation), 0],
                [0, -1j * cmath.sin(0.5 * self.rotation), cmath.cos(0.5 * self.rotation), 0],
                [1j * cmath.sin(0.5 * self.rotation), 0, 0, cmath.cos(0.5 * self.rotation)],
            ]
        )


class Rzz(BasicRotateGate):
    '''
    RotationZZ gate class
    Args:
        alpha: the angle to rotate
    '''
    def __init__(self, alpha):
        super().__init__()
        self.rotation = alpha

    def __str__(self):
        return "Rzz"

    def quqasm_str(self, qubits):
        return f"{self.__str__()}(" + str(self.rotation) + ") | (q[" + str(qubits[0].id) + "], q[" + str(qubits[1].id) + "])"

    def qasm_str(self, qubits):
        return "rzz (" + str(self.rotation) + ") q[" + str(qubits[0].id) + "],q[" + str(qubits[1].id) + "]"

    def __or__(self, qubits):
        '''
        Args:
            qubits: the quantum bits to aplly Rzz gate
        Example: Rzz(alpha) | (qr[0], qr[1])
        ''' 
        if len(qubits) != 2:
            raise AttributeError()
        targets = [q.id for q in qubits]
        cmd = Command(self, targets, rotation=[self.rotation], qasm=self.qasm_str(qubits), quqasm=self.quqasm_str(qubits))
        qubits[0].circuit.append_cmd(cmd)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix(
            [
                [cmath.exp(-0.5 * 1j * self.rotation), 0, 0, 0],
                [0, cmath.exp(0.5 * 1j * self.rotation), 0, 0],
                [0, 0, cmath.exp(0.5 * 1j * self.rotation), 0],
                [0, 0, 0, cmath.exp(-0.5 * 1j * self.rotation)],
            ]
        )


class SGate(BasicGate):
    '''Apply the single-qubit S gate'''
    def __init__(self):
        super().__init__()

    def __str__(self):
        return "S"
    
    def quqasm_str(self, qubit):
        return f"{self.__str__()} | q[" + str(qubit.id) + "]"

    def qasm_str(sel, qubit):
        return "s q[" + str(qubit.id) + "]"

    def __or__(self, qubit):
        '''
        Args:
            qubit: the quantum bit to aplly S gate
        Example: S | qr[0]
        ''' 
        targets = [qubit.id]
        cmd = Command(self, targets, qasm=self.qasm_str(qubit), quqasm=self.quqasm_str(qubit))
        qubit.circuit.append_cmd(cmd) 

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix([[1, 0], [0, 1j]])

S = SGate()

class SdgGate(BasicGate):
    '''Inverse of the SGate'''
    def __init__(self):
        super().__init__()

    def __str__(self):
        return "Sdg"

    def quqasm_str(self, qubit):
        return f"{self.__str__()} | q[" + str(qubit.id) + "]"

    def qasm_str(self, qubit):
        return "sdg q[" + str(qubit.id) + "]"

    def __or__(self, qubit):
        '''
        Args:
            qubit: the quantum bit to aplly Sdg gate
        Example: Sdg | qr[0]
        ''' 
        targets = [qubit.id]
        cmd = Command(self, targets, qasm=self.qasm_str(qubit), quqasm=self.quqasm_str(qubit))
        qubit.circuit.append_cmd(cmd) 

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix([[1, 0], [0, 1j]]).getH()

Sdg = SdgGate()

class TGate(BasicGate):
    '''Apply the single-qubit T gate.'''
    def __init__(self):
        super().__init__()

    def __str__(self):
        return "T"

    def quqasm_str(self, qubit):
        return f"{self.__str__()} | q[" + str(qubit.id) + "]"

    def qasm_str(self, qubit):
        return "t q[" + str(qubit.id) + "]"

    def __or__(self, qubit):
        '''
        Args:
            qubit: the quantum bit to aplly T gate
        Example: T | qr[0]
        ''' 
        targets = [qubit.id]
        cmd = Command(self, targets, qasm=self.qasm_str(qubit), quqasm=self.quqasm_str(qubit))
        qubit.circuit.append_cmd(cmd) 

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix([[1, 0], [0, cmath.exp(1j * cmath.pi / 4)]])

T = TGate()

class TdgGate(BasicGate):
    '''Inverse of the TGate'''
    def __init__(self):
        super().__init__()

    def __str__(self):
        return "Tdg"

    def quqasm_str(self, qubit):
        return f"{self.__str__()} | q[" + str(qubit.id) + "]"

    def qasm_str(self, qubit):
        return "tdg q[" + str(qubit.id) + "]"

    def __or__(self, qubit):
        '''
        Args:
            qubit: the quantum bit to aplly Tdg gate
        Example: Tdg | qr[0]
        ''' 
        targets = [qubit.id]
        cmd = Command(self, targets, qasm=self.qasm_str(qubit), quqasm=self.quqasm_str(qubit))
        qubit.circuit.append_cmd(cmd) 

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix([[1, 0], [0, cmath.exp(1j * cmath.pi / 4)]]).getH()

Tdg = TdgGate()

class SwapGate(BasicGate):
    '''Performs a SWAP gate between qubit1 and qubit2'''
    def __init__(self):
       super().__init__() 

    def __str__(self):
        return "Swap"

    def quqasm_str(self, qubits):
        return f"{self.__str__()} | (q[" + str(qubits[0].id) + "], q[" + str(qubits[1].id) + "])"

    def qasm_str(self, qubits):
        return "swap q[" + str(qubits[0].id) + "],q[" + str(qubits[1].id) + "]"

    def __or__(self, qubits):
        '''
        Args:
            qubits: qubits to swap
        Example: Swap | (qr[0], qr[1])
        ''' 
        targets = [q.id for q in qubits]
        cmd = Command(self, targets, qasm=self.qasm_str(qubits), quqasm=self.quqasm_str(qubits))
        qubits[0].circuit.append_cmd(cmd) 

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        # fmt: off
        return np.matrix([[1, 0, 0, 0],
                          [0, 0, 1, 0],
                          [0, 1, 0, 0],
                          [0, 0, 0, 1]])

Swap = SwapGate()

class SqrtSwapGate(BasicGate):
    '''Performs a sqrt SWAP gate between qubit1 and qubit2'''
    def __init__(self):
       super().__init__() 

    def __str__(self):
        return "SqrtSwap"

    def quqasm_str(self, qubits):
        return f"{self.__str__()} | (q[" + str(qubits[0].id) + "], q[" + str(qubits[1].id) + "])"

    def qasm_str(self, qubits):
        return "sqrtswap q[" + str(qubits[0].id) + "],q[" + str(qubits[1].id) + "]"

    def __or__(self, qubits):
        '''
        Args:
            qubits: qubits to swap
        Example: SqrtSwap | (qr[0], qr[1])
        ''' 
        targets = [q.id for q in qubits]
        cmd = Command(self, targets, qasm=self.qasm_str(qubits), quqasm=self.quqasm_str(qubits))
        qubits[0].circuit.append_cmd(cmd) 

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix(
            [
                [1, 0, 0, 0],
                [0, 0.5 + 0.5j, 0.5 - 0.5j, 0],
                [0, 0.5 - 0.5j, 0.5 + 0.5j, 0],
                [0, 0, 0, 1],
            ]
        )

SqrtSwap = SqrtSwapGate()

class SqrtXGate(BasicGate):
    '''Square-root X gate class.'''
    def __init__(self):
       super().__init__() 

    def __str__(self):
        return "SqrtX"

    def quqasm_str(self, qubit):
        return f"{self.__str__()} | q[" + str(qubit.id) + "]"

    def qasm_str(self, qubit):
        return "sx q[" + str(qubit.id) + "]"

    def __or__(self, qubit):
        '''
        Args:
            qubit: the quantum bit to aplly SqrtX gate
        Example: SqrtX | qr[0]
        '''  
        targets = [qubit.id]
        cmd = Command(self, targets, qasm=self.qasm_str(qubit), quqasm=self.quqasm_str(qubit))
        qubit.circuit.append_cmd(cmd) 

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return 0.5 * np.matrix([[1 + 1j, 1 - 1j], [1 - 1j, 1 + 1j]])

SqrtX = SqrtXGate()

class BarrierGate(BasicGate):
    """Barrier Gate"""
    def __init__(self):
       super().__init__() 

    def __str__(self):
        return "Barrier"

    def quqasm_str(self, qubits):
        target_str = ""
        target_cnt = len(qubits)
        for idx in range(target_cnt):
            target_str += "q[" + str(qubits[idx]) + "], "
        if target_str.endswith(", "):
            target_str = target_str[:-2]
        if target_cnt > 1:
            target_str = "(" + target_str + ")"
        return self.__str__() + " | " + target_str

    def qasm_str(self, qubits):
        target_str = ""
        target_cnt = len(qubits)
        for idx in range(target_cnt):
            target_str += "q[" + str(qubits[idx]) + "],"
        if target_str.endswith(","):
            target_str = target_str[:-1]
        return "barrier " + target_str

    def __or__(self, qubits):
        '''
        Args:
            qubits: the quantum bit to aplly Barrier gate
        Example: Barrier | (qr[0], qr[1])
        '''  
        templist = [qubits] if isinstance(qubits, Qubit) else qubits
        targets = [q.id for q in templist]
        cmd = Command(self, targets, qasm=self.qasm_str(targets), quqasm=self.quqasm_str(targets))
        templist[0].circuit.append_cmd(cmd) 

Barrier = BarrierGate()


class C(BasicGate):
    '''
    Meta operator, add the control bit to quantum gate,
    cooperate with other quantum gates,cannot be used alone
    Note: not all quantum gates have control bits, 
    currently supported: cx, cy, cz, cp, crx, cry, crz
    Args:
        gate: the gate wrap by control operator
        ctrl_cnt: number of control qubits
    '''
    def __init__(self, gate, ctrl_cnt=1):
        if isinstance(gate, (XGate, YGate, ZGate, P, Rx, Ry, Rz)):
            super().__init__() 
            self.gate = gate
            self.ctrl_cnt = ctrl_cnt
        else:
            raise AttributeError

    def __str__(self):
        return "C" + str(self.gate)

    def quqasm_str(self):
        controls = [ q.id for q in self.qubits[0:self.ctrl_cnt]]
        targets = [q.id for q in self.qubits[self.ctrl_cnt:]]
        cnt = len(controls)
        prefix = "C" * cnt + "(" + str(self.gate) + ")"
        if isinstance(self.gate, (XGate)):
            prefix = f"C(X, {cnt})"
        if isinstance(self.gate, (P, Rx, Ry, Rz)):
            prefix = f"C({self.gate}({self.gate.rotation}))"

        control_str = ""
        for idx in range(cnt):
            control_str += "q[" + str(controls[idx]) + "], "
        
        target_str = ""
        target_cnt = len(targets)
        for idx in range(target_cnt):
            target_str += "q[" + str(targets[idx]) + "], "

        return (prefix + " | " + "(" + control_str + target_str + ")").replace(', )', ')')
            
    def qasm_str(self):
        controls = [ q.id for q in self.qubits[0:self.ctrl_cnt]]
        targets = [q.id for q in self.qubits[self.ctrl_cnt:]]
        cnt = len(controls)
        prefix = "c" * cnt + str(self.gate).lower() + " "
        if isinstance(self.gate, (P, Rx, Ry, Rz)):
            prefix = f"c{str(self.gate).lower()} ({self.gate.rotation}) "
        control_str = ""
        for idx in range(cnt):
            control_str += "q[" + str(controls[idx]) + "],"
        
        target_str = ""
        target_cnt = len(targets)
        for idx in range(target_cnt):
            target_str += "q[" + str(targets[idx]) + "],"
        if target_str.endswith(","):
            target_str = target_str[:-1]
        return prefix + control_str + target_str

    def __or__(self, qubits):
        '''
        Args: 
            qubits: the left self.ctrl_cnt qubits are control qubits, the rest right bits are target qubits
        Example: C(Y) | (qr[0], qr[1]), C(X, 2) | (qr[0], qr[1], qr[2])
        '''
        self.qubits = qubits
        controls = [ q.id for q in qubits[0:self.ctrl_cnt]]
        targets = [q.id for q in qubits[self.ctrl_cnt:]]
        rotation = []
        if isinstance(self.gate, (P, Rx, Ry, Rz)):
            rotation = [self.gate.rotation]
        cmd = Command(self, targets, controls, rotation, qasm=self.qasm_str(), quqasm=self.quqasm_str())
        qubits[0].circuit.append_cmd(cmd)


CX = CNOT = C(X)
CY = C(Y)
CZ = C(Z)
Toffoli = C(X, 2)


class All(BasicGate):
    """
    Meta operator, provides unified operation of multiple qubits

    Args: 
        gate: the gate will apply to all qubits
    """
    def __init__(self, gate):
        self.gate = gate

    def __or__(self, qureg):
        """
        Args: 
            qureg: the qureg(represent a set of qubit) to apply gate
        Example: All(H) | qureg, All(Measure) | qureg
        """
        for q in qureg:
            self.gate | q
