#include <string>
#include <vector>
#include <map>
#include <math.h>
#include <time.h>
#include <fstream>
#include <memory>
#include <algorithm>

#include "pybind11/pybind11.h"
#include "pybind11/stl.h"
#include "QuEST/include/QuEST.h"

#ifndef M_PI 
#define M_PI 3.1415926535897932384626433832795028841971
#endif

using namespace std;
namespace py = pybind11;

using Qubits = vector<int>;
using Qurota = vector<double>;

typedef struct _Cmd {
    string gate;
    vector<int>  targets;
    vector<int>  controls;
    vector<double>  rotation;
    string desc;
} Cmd;

using Circuit = vector<Cmd>;

typedef struct _MeasureResult {
    // 量子比特id
    int id;
    // 测量结果
    int value;
} MeasureResult;

using MeasureSet = vector<MeasureResult>;

typedef struct _Outcome {
    // 比特位字符串 "000", "001"...
    string bitstr;
    // 比特位字符串出现的次数
    int count;
} Outcome;

using OutcomeSet = vector<Outcome>;

typedef struct _Result {
    MeasureSet measureSet;
    OutcomeSet outcomeSet;
} Result;


using GateFunc = void (*)(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation);

QuESTEnv g_env;
Qureg    g_qureg;
bool     g_bshowquantumgate = false;
Circuit  g_circuit;
Result   g_result;
int      g_initstate;
long long int g_initvalue;

extern "C" void invalidQuESTInputError(const char* errMsg, const char* errFunc) {
 
    string err = "in function " + string(errFunc) + ": " + string(errMsg);
    throw std::invalid_argument(err);
} 

void qh(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
    if (targets.size() != 1) {
        return;
    }
    
    hadamard(g_qureg, targets[0]);
}

void qp(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
    // matrix([[cmath.exp(1j * self.angle), 0], [0, cmath.exp(1j * self.angle)]]
    if (targets.size() != 1) {
        return;
    }

    double angle = rotation[0];
    ComplexMatrix2 m = {
        {{cos(angle), 0}, {0, cos(angle)}},
        {{sin(angle), 0}, {0, sin(angle)}}};

    applyMatrix2(g_qureg, targets[0], m);
}

void qr(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
    if (targets.size() != 1) {
        return;
    }

    phaseShift(g_qureg, targets[0], rotation[0]);
}

void qrx(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
    if (targets.size() != 1) {
        return;
    }

    rotateX(g_qureg, targets[0], rotation[0]);
}

void qry(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
    if (targets.size() != 1) {
        return;
    }

    rotateY(g_qureg, targets[0], rotation[0]);
}

void qrz(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
    if (targets.size() != 1) {
        return;
    }

    rotateZ(g_qureg, targets[0], rotation[0]);
}

void qrxx(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
    // projectq matrix
    // matrix(
    //         [
    //             [cmath.cos(0.5 * self.angle), 0, 0, -1j * cmath.sin(0.5 * self.angle)],
    //             [0, cmath.cos(0.5 * self.angle), -1j * cmath.sin(0.5 * self.angle), 0],
    //             [0, -1j * cmath.sin(0.5 * self.angle), cmath.cos(0.5 * self.angle), 0],
    //             [-1j * cmath.sin(0.5 * self.angle), 0, 0, cmath.cos(0.5 * self.angle)],
    //         ]
    //     )
    if (targets.size() != 2) {
        return;
    }

    ComplexMatrix4 m = {
            {{cos(0.5 * rotation[0]), 0, 0, 0},
            {0, cos(0.5 * rotation[0]), 0, 0},
            {0, 0, cos(0.5 * rotation[0]), 0},
            {0, 0, 0, cos(0.5 * rotation[0])}
        },
            {
            {0, 0, 0, -1 * sin(0.5 * rotation[0])},
            {0, 0, -1 * sin(0.5 * rotation[0]), 0},
            {0, -1 * sin(0.5 * rotation[0]), 0, 0},
            {-1 * sin(0.5 * rotation[0]), 0, 0, 0}
        }
    };

    applyMatrix4(g_qureg, targets[0], targets[1], m);
}

void qryy(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
    // projectq matrix
    // matrix(
    //     [
    //         [cmath.cos(0.5 * self.angle), 0, 0, 1j * cmath.sin(0.5 * self.angle)],
    //         [0, cmath.cos(0.5 * self.angle), -1j * cmath.sin(0.5 * self.angle), 0],
    //         [0, -1j * cmath.sin(0.5 * self.angle), cmath.cos(0.5 * self.angle), 0],
    //         [1j * cmath.sin(0.5 * self.angle), 0, 0, cmath.cos(0.5 * self.angle)],
    //     ]
    // )

    if (targets.size() != 2) {
        return;
    }

    ComplexMatrix4 m = {
            {{cos(0.5 * rotation[0]), 0, 0, 0},
            {0, cos(0.5 * rotation[0]), 0, 0},
            {0, 0, cos(0.5 * rotation[0]), 0},
            {0, 0, 0, cos(0.5 * rotation[0])}
        },
            {
            {0, 0, 0, 1 * sin(0.5 * rotation[0])},
            {0, 0, -1 * sin(0.5 * rotation[0]), 0},
            {0, -1 * sin(0.5 * rotation[0]), 0, 0},
            {1 * sin(0.5 * rotation[0]), 0, 0, 0}
        }
    };

    applyMatrix4(g_qureg, targets[0], targets[1], m);
}

void qrzz(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
    // projectq matrix
    // matrix(
    //     [
    //         [cmath.cos(0.5 * self.angle), 0, 0, 1j * cmath.sin(0.5 * self.angle)],
    //         [0, cmath.cos(0.5 * self.angle), -1j * cmath.sin(0.5 * self.angle), 0],
    //         [0, -1j * cmath.sin(0.5 * self.angle), cmath.cos(0.5 * self.angle), 0],
    //         [1j * cmath.sin(0.5 * self.angle), 0, 0, cmath.cos(0.5 * self.angle)],
    //     ]
    // )

    if (targets.size() != 2) {
        return;
    }

    double angle = rotation[0];
    ComplexMatrix4 m = {
        {
            {cos(0.5 * angle), 0, 0, 0},
            {0, cos(0.5 * angle), 0, 0},
            {0, 0, cos(0.5 * angle), 0},
            {0, 0, 0, cos(0.5 * angle)}
        },
        {
            {-1 * sin(0.5 * angle), 0, 0, 0},
            {0, sin(0.5 * angle), 0, 0},
            {0, 0, sin(0.5 * angle), 0},
            {0, 0, 0, -1 * sin(0.5 * angle)}
        }
    };

    applyMatrix4(g_qureg, targets[0], targets[1], m);
}

void qx(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
    if (targets.size() != 1) {
        return;
    }

    pauliX(g_qureg, targets[0]);
}

void qy(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
    if (targets.size() != 1) {
        return;
    }

    pauliY(g_qureg, targets[0]);
}

void qz(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
    if (targets.size() != 1) {
        return;
    }

    pauliZ(g_qureg, targets[0]);
}

void qs(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
    if (targets.size() != 1) {
        return;
    }

    sGate(g_qureg, targets[0]);
}

void qt(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
    if (targets.size() != 1) {
        return;
    }

    tGate(g_qureg, targets[0]);
}

void qsdg(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
     if (targets.size() != 1) {
        return;
    }

    // sgate matrix([[1, 0], [0, 1j]])
    // 共轭转置：matrix([[1, 0], [0, -1j]])
    ComplexMatrix2 m = {
        {
            {1, 0},
            {0, 0}
        },
        {
            {0, 0},
            {0, -1}
        }
    };

    applyMatrix2(g_qureg, targets[0], m); 
}

void qtdg(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
     if (targets.size() != 1) {
        return;
    }

    // matrix([[1, 0], [0, cmath.exp(1j * cmath.pi / 4)]])
    // 共轭转置：matrix([[1, 0], [0, cmath.exp(-1j * cmath.pi / 4)]])
    ComplexMatrix2 m = {
        {
            {1, 0},
            {0, cos(M_PI / 4)}
        },
        {
            {0, 0},
            {0, -sin(M_PI / 4)}
        }
    };

    applyMatrix2(g_qureg, targets[0], m); 
}

void qsqrtx(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
    if (targets.size() != 1) {
        return;
    }

    // projectq matrix
    // 0.5 * np.matrix([[1 + 1j, 1 - 1j], [1 - 1j, 1 + 1j]])
    ComplexMatrix2 m = {
        {
            {0.5, 0.5},
            {0.5, 0.5}
        },
        {
            {0.5, -0.5},
            {-0.5, 0.5}
        }
    };

    applyMatrix2(g_qureg, targets[0], m);
}

void qsqrtswap(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
    if (targets.size() != 2) {
        return;
    }

    sqrtSwapGate(g_qureg, targets[0], targets[1]);
}

void qswap(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
   if (targets.size() != 2) {
        return;
    }

    swapGate(g_qureg, targets[0], targets[1]);
}

void qcx(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
    if (targets.size() != 1 || controls.size() != 1)
        return;

    controlledNot(g_qureg, controls[0], targets[0]);
}

void qcy(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
    if (targets.size() != 1 || controls.size() != 1) {
        return;
    } 

    controlledPauliY(g_qureg, controls[0], targets[0]);
}

void qcz(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
   // cmd.controls.insert(cmd.controls.cend(), cmd.targets.cbegin(), cmd.targets.cend());
    vector<int32_t> tempCtls;
    tempCtls.assign(controls.begin(), controls.end());
    tempCtls.insert(tempCtls.end(), targets.begin(), targets.end());

    auto ctrls = make_unique<int[]>(tempCtls.size());
    copy(tempCtls.begin(), tempCtls.end(), ctrls.get());
    
    multiControlledPhaseFlip(g_qureg, 
        ctrls.get(), 
        tempCtls.size()
    );
}

void qu3(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
   if (rotation.size() != 3 || targets.size() != 1) {
        return;
    }

    auto alpha = rotation[0];
    auto beta = rotation[1];
    auto theta = rotation[2];
     ComplexMatrix2 m = {
        {
            {cos(alpha / 2), -1 * sin(alpha / 2) * cos(theta)},
            {sin(alpha / 2) * cos(beta), cos(alpha / 2) * cos(beta + theta)}
        },
        {
            {0, -1 * sin(theta) * sin(alpha / 2)},
            {sin(beta) * sin(alpha / 2), sin(beta + theta) * cos(alpha / 2)}
        }
    };
    unitary(g_qureg, targets[0], m);
}

void qu2(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
   if (rotation.size() != 2 || targets.size() != 1) {
        return;
    }

    auto alpha = rotation[0];
    auto beta = rotation[1];
    double factor = 1 / sqrt(2);
    ComplexMatrix2 m = {
        {
            {1 * factor, -factor * cos(beta)},
            {factor * cos(alpha), factor * cos(alpha + beta)}
        },
        {
            {0, -factor * sin(beta)},
            {factor * sin(alpha), factor * sin(alpha + beta)}
        }
    };
    unitary(g_qureg, targets[0], m);
}

void qu1(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation)
{
   if (rotation.size() != 1 || targets.size() != 1) {
        return;
    }

    auto alpha = rotation[0];
    ComplexMatrix2 m = {
        {
            {1, 0},
            {0, cos(alpha)}
        },
        {
            {0, 0},
            {0, sin(alpha)}
        }
    };
    unitary(g_qureg, targets[0], m);
}


void qcrx(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation) {
    if (rotation.size() != 1 || targets.size() != 1 || controls.size() != 1) {
        return;
    }

    controlledRotateX(g_qureg, controls[0], targets[0], rotation[0]);
}

void qcry(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation) {
    if (rotation.size() != 1 || targets.size() != 1 || controls.size() != 1) {
        return;
    }

    controlledRotateY(g_qureg, controls[0], targets[0], rotation[0]);
}

void qcrz(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation) {
    if (rotation.size() != 1 || targets.size() != 1 || controls.size() != 1) {
        return;
    }

    controlledRotateZ(g_qureg, controls[0], targets[0], rotation[0]);
}

int qmeasure(const int qubit)
{
    return measure(g_qureg, qubit);
}

map<string, GateFunc> g_gateMap = {
    make_pair<string, GateFunc>("H", &qh),
    make_pair<string, GateFunc>("P", &qp),
    make_pair<string, GateFunc>("R", &qr),
    make_pair<string, GateFunc>("Rx", &qrx), 
    make_pair<string, GateFunc>("Ry", &qry),
    make_pair<string, GateFunc>("Rz", &qrz),
    make_pair<string, GateFunc>("Rxx", &qrxx),
    make_pair<string, GateFunc>("Ryy", &qryy),
    make_pair<string, GateFunc>("Rzz", &qrzz),
    make_pair<string, GateFunc>("NOT", &qx),
    make_pair<string, GateFunc>("X", &qx),
    make_pair<string, GateFunc>("Y", &qy),
    make_pair<string, GateFunc>("Z", &qz),
    make_pair<string, GateFunc>("S", &qs),
    make_pair<string, GateFunc>("T", &qt),
    make_pair<string, GateFunc>("Sdg", &qsdg),
    make_pair<string, GateFunc>("Tdg", &qtdg),
    make_pair<string, GateFunc>("SqrtX", &qsqrtx),
    make_pair<string, GateFunc>("SqrtSwap", &qsqrtswap),
    make_pair<string, GateFunc>("Swap", &qswap),
    make_pair<string, GateFunc>("CNOT", &qcx),
    make_pair<string, GateFunc>("CX", &qcx),
    make_pair<string, GateFunc>("CY", &qcy),
    make_pair<string, GateFunc>("CZ", &qcz),
    make_pair<string, GateFunc>("U3", &qu3),
    make_pair<string, GateFunc>("U2", &qu2),
    make_pair<string, GateFunc>("U1", &qu1),
    make_pair<string, GateFunc>("CRx", &qcrx),
    make_pair<string, GateFunc>("CRy", &qcry),
    make_pair<string, GateFunc>("CRz", &qcrz),
};

void init(int qubits, int state, long long int value, bool showquantumgate) {
    g_env = createQuESTEnv();
    g_qureg = createQureg(qubits, g_env);
    g_bshowquantumgate = showquantumgate;
    g_circuit.clear();
    g_result.measureSet.clear();
    g_result.outcomeSet.clear();

    g_initstate = state;
    g_initvalue = value;

    if (g_bshowquantumgate) {
        startRecordingQASM(g_qureg);
        cout << "################# qutrunk Cmd Start! #################" << endl;
        cout << "qreg q[" << qubits << "]" << endl;
        cout << "creg c[" << qubits << "]" << endl;
    }

    if (state == 2) {
        initPlusState(g_qureg);
    } else if (state == 3) {
        initClassicalState(g_qureg, value);
    }
}

void release() {
    if (g_bshowquantumgate) {
        cout << "################# qutrunk Cmd End! ###################" << endl;
        cout << "***************** QuEST Cmd Start! *****************" << endl;
        printRecordedQASM(g_qureg);
        cout << "***************** QuEST Cmd End! *****************" << endl;
        clearRecordedQASM(g_qureg);
        stopRecordingQASM(g_qureg);
    }
    destroyQureg(g_qureg, g_env);
    destroyQuESTEnv(g_env);
}

// todo: 直接传cmd
int execCmd(const string& gate, 
    const Qubits& targets,
    const Qubits& controls,
    const Qurota& rotation,
    const string& desc)
{
    if (gate == "Measure") {
        int res = qmeasure(targets[0]);
        MeasureResult mr{targets[0], res};
        g_result.measureSet.push_back(mr);
    }

    const auto it = g_gateMap.find(gate);
    if (it == g_gateMap.end()) {
        return -1;
    }

    if (g_bshowquantumgate) {
        cout << desc << endl;
    }

    try
    {
        (*(it->second))(gate, targets, controls, rotation);
    }
    catch(invalid_argument& ex)
    {
        std::cout << "(" << gate << ")" << " Invalid argument exception " << ex.what();
        return -1;
    }
    catch(exception& ex)
    {
        std::cout << "(" << gate << ")" << " Unknown exception " << ex.what();
        return -1;
    }
    
    return 0;
}

void pack_measrue_result() {
    sort(g_result.measureSet.begin(), g_result.measureSet.end(), [](MeasureResult& a, MeasureResult& b) {return a.id < b.id;});
    string bitstr = "";
    for (auto &m : g_result.measureSet) {
        int val = m.value;
        bitstr.append(to_string(val));
    }

    int idx = -1;
    for (int i = 0; i < g_result.outcomeSet.size(); ++i) {
        if (g_result.outcomeSet[i].bitstr == bitstr) {
            idx = i;
            break; 
        }
    }
    if (idx >= 0) {
        g_result.outcomeSet[idx].count += 1;
    } else {
        Outcome out{bitstr, 1};
        g_result.outcomeSet.push_back(out);
    }
}

void send_circuit(const Circuit& circuit, const bool final) {
    for (const auto & cmd : circuit) {
        // 缓存指令
        g_circuit.push_back(cmd);
        // 执行指令
        execCmd(cmd.gate, cmd.targets, cmd.controls, cmd.rotation, cmd.desc);
    }

    if (final) {
        pack_measrue_result();
    }
}

Result run(int shots) {
    // 如果只运行一次，那么在线路发送完成后已经自动运行一次了
    int run_times = shots - 1;
    while (run_times > 0) {
        g_result.measureSet.clear();
        
        // 每次运行需要重置状态
        if (g_initstate == 2) {
            initPlusState(g_qureg);
        } else if (g_initstate == 3) {
            initClassicalState(g_qureg, g_initvalue);
        } else {
            initZeroState(g_qureg);
        }

        for (const auto & cmd : g_circuit) {
            // 执行指令
            execCmd(cmd.gate, cmd.targets, cmd.controls, cmd.rotation, cmd.desc);
        }
        pack_measrue_result();
        run_times -= 1;
    }

    release();

    sort(g_result.outcomeSet.begin(), g_result.outcomeSet.end(), [](Outcome& a, Outcome& b) {return a.bitstr.compare(b.bitstr) < 0;});

    return g_result;
}

double getProbOfAmp(const int64_t index) {
    return getProbAmp(g_qureg, index);
}

double getProbOfOutcome(const int32_t qubit, const int32_t outcome) {
    return calcProbOfOutcome(g_qureg, qubit, outcome);
}

std::vector<double> getProbOfAllOutcome(const std::vector<int32_t> & qubits) {
    std::vector<double> qalloutcome;

    int numQubits = qubits.size();
    if (numQubits == 0) {
        return qalloutcome;
    }

    int numOutcomes = 1LL << numQubits;

    auto probs = make_unique<qreal[]>(numOutcomes);
    auto bits = make_unique<int[]>(numOutcomes);

    copy(qubits.begin(), qubits.end(), bits.get());
    calcProbOfAllOutcomes(probs.get(), g_qureg, bits.get(), numQubits);

    for (auto i = 0; i < numOutcomes; ++i) {
        qalloutcome.push_back(probs[i]);
    }

    return qalloutcome;
}

std::vector<std::string> getAllState() {
    std::vector<std::string> qallstate;

    reportState(g_qureg);

    char filename[100];
    sprintf(filename, "state_rank_%d.csv", g_qureg.chunkId);
    fstream statefile;
	statefile.open(filename, ifstream::in);
	if (statefile.is_open())
	{
		string buff;
		while (getline(statefile, buff))
		{
			qallstate.push_back(buff);
		}
    } 
    statefile.close();

    return qallstate;
}

void apply_QFT(const std::vector<int32_t> & qubits) {
    if (qubits.size() == 0) {
        return ;
    }

    auto bits = std::make_unique<int32_t []>(qubits.size());

    copy(qubits.begin(), qubits.end(), bits.get());
    
    applyQFT(g_qureg, bits.get(), qubits.size());
}

void apply_Full_QFT() {
    applyFullQFT(g_qureg);
}



PYBIND11_MODULE(simulator, m) {
    m.doc() = "simulator plugin"; // optional module docstring

    m.def("init", &init, "init sim envirement", py::arg("qubits"), py::arg("state"), py::arg("value"), py::arg("showquantumgate"));
    // m.def("release", &release, "release resource");
    // m.def("execCmd", &execCmd, "execute circuit", py::arg("gate"), py::arg("targets"), py::arg("controls"), py::arg("rotation"), py::arg("desc"));
    m.def("getProbOfAmp", &getProbOfAmp, "getProbOfAmp", py::arg("index"));
    m.def("getProbOfOutcome", &getProbOfOutcome, "getProbOfOutcome", py::arg("qubit"), py::arg("outcome"));
    m.def("getProbOfAllOutcome", &getProbOfAllOutcome, "getProbOfAllOutcome", py::arg("qubits"));
    m.def("getAllState", &getAllState, "getAllState");
    m.def("apply_QFT", &apply_QFT, "apply_QFT", py::arg("qubits"));
    m.def("apply_Full_QFT", &apply_Full_QFT, "apply_Full_QFT");

    py::class_<Cmd>(m, "Cmd")
        .def(py::init())
        .def_readwrite("gate", &Cmd::gate)
        .def_readwrite("targets", &Cmd::targets)
        .def_readwrite("controls", &Cmd::controls)
        .def_readwrite("rotation", &Cmd::rotation)
        .def_readwrite("desc", &Cmd::desc);

    m.def("send_circuit", &send_circuit, "send_circuit", py::arg("circuit"), py::arg("final"));
    m.def("run", &run, "run", py::arg("shots"));

    py::class_<MeasureResult>(m, "MeasureResult")
        .def(py::init())
        .def_readwrite("id", &MeasureResult::id)
        .def_readwrite("value", &MeasureResult::value);

    py::class_<Outcome>(m, "Outcome")
        .def(py::init())
        .def_readwrite("bitstr", &Outcome::bitstr)
        .def_readwrite("count", &Outcome::count);

    py::class_<Result>(m, "Result")
        .def(py::init())
        .def_readwrite("measureSet", &Result::measureSet)
        .def_readwrite("outcomeSet", &Result::outcomeSet);
}